#ifndef LR1_DATA_H
#define LR1_DATA_H

#include <iostream>
#include <vector>
#include <array>

struct Node {
    int node_id = 0;
    static constexpr int dimension = 3;
    std::array<double, dimension> xyz = {0, 0, 0};
    bool flag_vertex = 0;

    Node() = default;
    Node(int, const std::array<double, dimension>&, bool);
    Node(const Node&);

    friend std::istream& operator>>(std::istream&, Node&);
    friend std::ostream& operator<<(std::ostream&, const Node&);
};

struct FiniteElement {
    int f_elem_id;
    int material_id;
    std::vector<int> nodes_id;

    FiniteElement() = default;
    FiniteElement(int, int, const std::vector<int>&);

    friend std::istream& operator>>(std::istream&, FiniteElement&);
    friend std::ostream& operator<<(std::ostream&, const FiniteElement&);
};

struct BoundaryFiniteElement {
    int bf_elem_id;
    int boundary_id;
    std::vector<int> nodes_id;

    BoundaryFiniteElement() = default;
    BoundaryFiniteElement(int, int, const std::vector<int>&);

    friend std::istream& operator>>(std::istream&, BoundaryFiniteElement&);
    friend std::ostream& operator<<(std::ostream&, const BoundaryFiniteElement&);
};

struct Edge {
    std::pair<int, int> nodes_of_edge;
    int middle_node = -1;

    Edge() = default;
    Edge(int, int);

    void update_mid(int);

    Edge& operator=(const Edge&);
    bool operator==(const Edge&) const;

    friend std::istream& operator>>(std::istream&, Edge&);
    friend std::ostream& operator<<(std::ostream&, Edge&);
};


#endif //LR1_DATA_H