#include "MeshLoader.h"
#include "Hash.h"
#include <unordered_map>
#include <unordered_set>
#include <algorithm>

const std::vector<Node>& MeshLoader::get_nodes() const {
    return node;
}

const std::vector<FiniteElement>& MeshLoader::get_f_elems() const {
    return f_elem;
}

const std::vector<BoundaryFiniteElement>& MeshLoader::Get_bf_elems() const {
    return bf_elem;
}

std::vector<int> MeshLoader::find_f_elem_by_nodes(int id1, int id2, int id3) const {
    std::vector<int> found_elems;
    auto temp = f_elem.begin();
    while (temp != f_elem.end()) {
        temp = std::find_if(temp, f_elem.end(), [id1, id2, id3](const FiniteElement& el) {
            return std::find(el.nodes_id.begin(), el.nodes_id.end(), id1) != el.nodes_id.end()
                   && std::find(el.nodes_id.begin(), el.nodes_id.end(), id2) != el.nodes_id.end()
                   && std::find(el.nodes_id.begin(), el.nodes_id.end(), id3) != el.nodes_id.end();
        });
        if (temp != f_elem.end()) {
            found_elems.push_back(temp->f_elem_id);
            ++temp;
        }
    }
    return found_elems;
}

std::vector<int> MeshLoader::find_f_elem_by_edge(int id1, int id2) const {
    std::vector<int> found_elems;
    auto temp = f_elem.begin();
    while (temp != f_elem.end()) {
        temp = std::find_if(temp, f_elem.end(), [id1, id2](const FiniteElement& el) {
            return std::find(el.nodes_id.begin(), el.nodes_id.end(), id1) != el.nodes_id.end() &&
                   std::find(el.nodes_id.begin(), el.nodes_id.end(), id2) != el.nodes_id.end();
        });
        if (temp != f_elem.end()) {
            found_elems.push_back(temp->f_elem_id);
            ++temp;
        }
    }
    return found_elems;
}

std::set <int> MeshLoader::get_nodes_by_boundary(int id) const{
    std::set <int> found_nodes;
    auto temp = bf_elem.begin();
    while (temp != bf_elem.end()) {
        temp = std::find_if(temp, bf_elem.end(), [id](const BoundaryFiniteElement& n) {
            return n.boundary_id == id;
        });
        if (temp != bf_elem.end()) {
            for (const auto& n_in_boundary: temp->nodes_id) {
                found_nodes.insert(n_in_boundary);
            }
            ++temp;
        }
    }
    return found_nodes;
}

std::set <int> MeshLoader::get_f_elems_by_material(int material_id) const {
    std::set <int> found_elems;
    auto temp = f_elem.begin();
    while (temp != f_elem.end()) {
        temp = std::find_if(temp, f_elem.end(), [material_id](const FiniteElement& f) {
            return f.material_id == material_id;
        });
        if (temp != f_elem.end()) {
            found_elems.insert(temp->f_elem_id);
            ++temp;
        }
    }
    return found_elems;
}

std::set <int> MeshLoader::get_bf_elems_by_boundary(int id) const {
    std::set <int> found_elems;
    auto temp = bf_elem.begin();
    while (temp != bf_elem.end()) {
        temp = std::find_if(temp, bf_elem.end(), [id](const BoundaryFiniteElement& bf) {
            return bf.boundary_id == id;
        });
        if (temp != bf_elem.end()) {
            found_elems.insert(temp->bf_elem_id);
            ++temp;
        }
    }
    return found_elems;
}

void MeshLoader::insertion_in_middle(){
    std::unordered_set <Edge, hash> considered_edges;
    for (auto& el : f_elem) {
        std::vector<int> smth_nodes_id = el.nodes_id;
        for (auto first_n = 0; first_n < 4; ++first_n) {
            for (auto second_n = first_n + 1; second_n < 4; ++second_n) {
                Edge smth_edge(smth_nodes_id[first_n], smth_nodes_id[second_n]);
                if (considered_edges.insert(smth_edge).second) {
                    Node new_node = get_middle_node(smth_edge);
                    smth_edge.update_mid(new_node.node_id);
                    node.push_back(new_node);
                    el.nodes_id.push_back(new_node.node_id);
                }
                else {
                    el.nodes_id.push_back(smth_edge.middle_node);
                }
            }
        }
    }for (auto& el : bf_elem) {
        std::vector<int> smth_nodes_id = el.nodes_id;
        for (auto first_n = 0; first_n < 3; ++first_n) {
            for (auto second_n = first_n + 1; second_n < 3; ++second_n) {
                Edge smth_edge(smth_nodes_id[first_n], smth_nodes_id[second_n]);
                for (auto second_n = first_n + 1; second_n < 3; ++second_n) {
                    auto found_edge = considered_edges.find({ smth_nodes_id[first_n], smth_nodes_id[second_n] });
                    if (found_edge == considered_edges.end()) {
                        found_edge = considered_edges.find({ smth_nodes_id[second_n], smth_nodes_id[first_n] });
                    }
                    el.nodes_id.push_back(found_edge->middle_node);
                }
            }
        }
    }
}


std::array<double, 3> MeshLoader::get_middle_of_edge(const Edge& p_edge) {
    return
            { ((node.at(p_edge.nodes_of_edge.first - 1).xyz.at(0)) + (node.at(p_edge.nodes_of_edge.second - 1).xyz.at(0))) / 2
                    , ((node.at(p_edge.nodes_of_edge.first - 1).xyz.at(1)) + (node.at(p_edge.nodes_of_edge.second - 1).xyz.at(1))) / 2
                    , ((node.at(p_edge.nodes_of_edge.first - 1).xyz.at(2)) + (node.at(p_edge.nodes_of_edge.second - 1).xyz.at(2))) / 2 };
}

Node MeshLoader::get_middle_node(const Edge& smth_edge) {
    Node new_node(node.size() + 1, get_middle_of_edge(smth_edge), 0);
    return new_node;
}


Node& MeshLoader::get_node_by_id(int p_id) {
    return node.at(p_id);
}

std::array<double, 3> MeshLoader::get_xyz_of_node_by_id(int p_id) {
    return node.at(p_id).xyz;
}

std::vector<std::set<int>> MeshLoader::get_neighbor_nodes() const {
    std::vector<std::set<int>> neighbors(node.size() + 1);
    for (const auto& new_elem : f_elem) {
        for (auto new_node_id : new_elem.nodes_id)
            for (auto adj_node_id : new_elem.nodes_id)
                if (new_node_id != adj_node_id)
                    neighbors[new_node_id].insert(adj_node_id);
    }

    return neighbors;
}

void MeshLoader::print_node(const Node &node) {
    std::cout << node << std::endl;
}

void MeshLoader::print_f_elem(const FiniteElement &f_elem) {
    std::cout << f_elem << std::endl;
}

void MeshLoader::print_bf_elem(const BoundaryFiniteElement &bf_elem) {
    std::cout << bf_elem << std::endl;
}